/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2007-2009 Bull S.A.S.
 * Contact: cmi@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: CMIInvocationHandler.java 2274 2009-03-29 20:12:55Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.rpc;

import java.io.Serializable;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.jcip.annotations.ThreadSafe;

import org.ow2.cmi.controller.client.ClientClusterViewManager;
import org.ow2.cmi.controller.common.ClusterViewManager;
import org.ow2.cmi.controller.factory.ClusterViewManagerFactory;
import org.ow2.cmi.controller.provider.ClientClusterViewProvider;
import org.ow2.cmi.event.CMIEventComponent;
import org.ow2.cmi.event.CMIManagedObject;
import org.ow2.cmi.lb.NoLoadBalanceableException;
import org.ow2.cmi.lb.decision.DecisionManager;
import org.ow2.cmi.lb.decision.DecisionManager.Decision;
import org.ow2.cmi.lb.policy.IPolicy;
import org.ow2.cmi.reference.CMIProxyHandle;
import org.ow2.cmi.reference.CMIReference;
import org.ow2.cmi.reference.CMIReferenceable;
import org.ow2.cmi.reference.ObjectNotFoundException;
import org.ow2.util.component.api.Component;
import org.ow2.util.event.api.IEventDispatcher;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * An abstract implementation of the interface {@link InvocationHandler} allowing the choice of the mandatory to perform an invocation.
 * @param <T> the type of the remote object.
 * @author Loris Bouzonnet
 */
@ThreadSafe
public abstract class CMIInvocationHandler<T> implements InvocationHandler, Serializable, CMIManagedObject {

    /**
     * Id for serializable class.
     */
    private static final long serialVersionUID = -2257434302509492634L;

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(CMIInvocationHandler.class);

    /**
     * The classloader to use for the invocation.
     */
    private transient ClassLoader classLoader;

    /**
     * A CMI handler is relative to a object.
     */
    protected final String objectName;

    /**
     * A CMI handler is relative to a protocol.
     */
    protected final String protocolName;

    /**
     * An instance of ClientClusterViewManager to access to the cluster view.
     */
    protected transient ClusterViewManager clusterViewManager;

    /**
     * The object to use for the invocation.
     */
    private CMIReferenceable<T> currentRef;

    /**
     * Interface of the object.
     */
    protected transient Class<? extends T> itf;

    /**
     * True if the previous object must be reused.
     */
    private boolean keepCurrentRef;

    /**
     * Keep a trace of the handle in order to update the HTTP session that reference it.
     */
    protected transient CMIProxyHandle cmiProxyHandle = null;

    /**
     * Duration of the last invocation.
     */
    private long lastInvocDuration;

    private transient IEventDispatcher dispatcher;

    /**
     * The provider id.
     */
    private transient String cmiManagedObjectId;

    private static transient long id = 0;

    /**
     * Construct an invocation handler allowing the choice of the mandatory to perform an invocation.
     * @param clusterViewManager an instance of ClientClusterViewManager to access to the cluster view
     * @param objectName the name of the object on which performing invocations
     * @param protocolName the protocol used to perform invocations
     * @param keepCurrentRef true if the previous object must be reused
     * @param itf
     */
    protected CMIInvocationHandler(
            final ClusterViewManager clusterViewManager,
            final String objectName,
            final String protocolName,
            final boolean keepCurrentRef,
            final Class<? extends T> itf) {
        this.clusterViewManager = clusterViewManager;
        this.protocolName = protocolName;
        this.objectName = objectName;
        this.keepCurrentRef = keepCurrentRef;
        this.itf = itf;
    }

    /**
     * Set the classloader to use for the invocations.
     * @param classLoader the classloader to use for the invocations
     */
    protected void setClassLoader(final ClassLoader classLoader) {
        this.classLoader = classLoader;
    }


    /**
     * Get a reference to the first component matching the interface.
     * @param <Comp> The interface type.
     * @param itf The interface class.
     * @return The component.
     */
    public <Comp extends Component> Comp getComponent(final Class<Comp> itf) {
        return clusterViewManager.getComponent(itf);
    }

    /**
     * @return
     */
    public String getCMIManagedObjectId() {
        return cmiManagedObjectId;

    }
    /**
     * Invoke a method through the CMI proxy.
     * @param proxy The proxy to use.
     * @param method The method to call.
     * @param args The arguments of the method.
     * @return the value to return from the method invocation on the proxy instance
     * @throws Throwable
     */
    @SuppressWarnings("unchecked")
    public Object invoke(final Object proxy, final Method method, final Object[] args) throws Throwable {

        if(clusterViewManager == null) {
            clusterViewManager =
                ClusterViewManagerFactory.getFactory().getClusterViewManager();
        }
        if(classLoader == null) {
            classLoader = Thread.currentThread().getContextClassLoader();
        }
        if(itf == null) {
            clusterViewManager.addObjectToWatch(objectName);
            try {
                itf = (Class<T>) clusterViewManager.getInterface(objectName);
            } catch (ObjectNotFoundException e) {
                logger.error("Cannot retrieve the interface for object with name " + objectName, e);
                throw new CMIInvocationHandlerException(
                        "Cannot retrieve the interface for object with name " + objectName, e);
            }
        }
        if(cmiManagedObjectId == null) {
            // Register to event component.
            registerCMIManagedObject();
        }
        checkInitialized();

        if (method.getDeclaringClass() == Object.class) {
            return invokeObjectMethod(proxy, method, args);
        }
        String methodName = method.getName();
        if(methodName.endsWith("_CMI")) {
            return invokeCMIProxy(
                    methodName.substring(0, methodName.length() - 4), proxy);
        }
        return invokeRemoteMethod(proxy, method, args);
    }

    private void registerCMIManagedObject() {
        dispatcher = clusterViewManager.getEventDispatcher();
        if(dispatcher != null) {
            cmiManagedObjectId = "/cmi/" + objectName + "/" + protocolName + "/" + id++;

            // Register to event component.
            CMIEventComponent eventComponent = getComponent(CMIEventComponent.class);
            if (eventComponent != null) {
                eventComponent.registerCMIManagedObject(this, this.dispatcher);
            }
        }
    }

    /**
     * Handle the invocation of methods defined in the interface {@link CMIProxy}.
     * @param methodName the method name without the suffix _CMI
     * @param proxy the proxy to use
     * @return the result of invocation
     * @throws CMIInvocationHandlerException
     */
    private Object invokeCMIProxy(final String methodName, final Object proxy) throws CMIInvocationHandlerException {
        if(methodName.equals("getObjectName")) {
            return objectName;
        } else if(methodName.equals("getInterface")) {
            return itf;
        } else if(methodName.equals("getProtocolName")) {
            return protocolName;
        } else if(methodName.equals(getHandleMethodName())) {
            return getHandle((CMIProxy) proxy);
        } else if(methodName.equals("getCurrentCMIRef")) {
            return currentRef;
        } else if(methodName.equals("isCMIRefKept")) {
            return keepCurrentRef;
        }
        throw new CMIInvocationHandlerException(
                "CMIProxy doesn't define the method " + methodName);
    }

    /**
     * Because a {@link CMIProxy} can be serialized, this method restores lazily the transient fields.
     * @throws CMIInvocationHandlerException if some fields cannot be retrieved
     */
    protected abstract void checkInitialized() throws CMIInvocationHandlerException;

    /**
     * Handles java.lang.Object methods.
     **/
    private Object invokeObjectMethod(final Object proxy, final Method method, final Object... args) {
        String name = method.getName();

        if (name.equals("hashCode")) {
            return Integer.valueOf(hashCode());

        } else if (name.equals("equals")) {
            Object obj = args[0];
            boolean b =
                proxy == obj
                || (obj != null
                        && Proxy.isProxyClass(obj.getClass())
                        && equals(Proxy.getInvocationHandler(obj)));
            return Boolean.valueOf(b);

        } else if (name.equals("toString")) {
            return proxyToString(proxy);

        } else {
            logger.error("unexpected Object method: {0}", method);
            throw new IllegalArgumentException("unexpected Object method: " + method);
        }
    }

    /**
     * @param proxy
     * @return a string representation for a proxy that uses this invocation handler.
     **/
    protected String proxyToString(final Object proxy) {
        return "object:" + objectName
        + ", protocol:" + protocolName
        + ", object:" + currentRef
        + ", keepCurrentRef:" + keepCurrentRef
        + ", itf:" + itf;
    }

    protected Object invokeRemoteMethod(
            final Object proxy, final Method method, final Object... args) throws Throwable {
        String methodName = method.getName();
        List<CMIReference> cmiReferences = null;
        IPolicy<CMIReference> policy = null;

        // Retrieve required data
        try {
            policy = clusterViewManager.getPolicy(objectName);
            logger.debug("Object: {0} - Invoke: {1} > Policy: {2}", objectName, methodName, policy);
            cmiReferences = new ArrayList<CMIReference>(clusterViewManager.getCMIReferences(objectName, protocolName));
            logger.debug("Object: {0} - Invoke:{1} > CMIReferences: {2}", objectName, methodName, cmiReferences);
        } catch (Exception e) {
            logger.error("Object: {0} - Invoke:{1} > Cannot retrieve data", objectName, method, e);
            throw new CMIInvocationHandlerException(
                    "Object: " + objectName + " - Invoke: " + methodName + " > Cannot retrieve data", e);
        }

        CMIReference cmiReference;
        CMIReferenceable<T> cmiReferenceable;
        DecisionManager<Object> decisionManagerOnReturn = null;

        // While the invocation is not done and some mandatories are still available, try to invoke...
        do {
            cmiReference = null;
            cmiReferenceable = null;

            // Check if a previous mandatory should be reused
            if(keepCurrentRef && currentRef != null) {
                cmiReference = currentRef.getReference();
                cmiReferenceable = currentRef;
            }
            if(cmiReferenceable == null) {
                if(cmiReferences.isEmpty()) {
                    logger.error("Object:{0} - Invoke: {1} > No server available! Skipping the invocation...",
                            objectName, methodName);
                    throw new CMIInvocationHandlerException("Object: " + objectName + " - Invoke: " + methodName
                            + " > No server available! Skipping the invocation...");
                }

                // Choose a new mandatory
                logger.debug("Object:{0} - Invoke: {1} > Choosing a node...", objectName, methodName);
                try {
                    cmiReference = policy.choose(cmiReferences);
                } catch (NoLoadBalanceableException e) {
                    logger.error("Object: {0} - Invoke: {1} > Cannot get a CMIReference for protocol {2}",
                            objectName, methodName, protocolName, e);
                    throw new CMIInvocationHandlerException(
                            "Object: " + objectName + " - Invoke: " + methodName
                            + " > Cannot get a CMIReference for protocol " + protocolName, e);
                }
                logger.debug("Object: {0} - Invoke: {1} > Trying to get a CMIReferenceable for {2}...",
                        objectName, methodName, cmiReference);

                DecisionManager<CMIReferenceable<T>> decisionManagerOnChoose;
                try {
                    // Get the CMIReferenceable stating for a proxy or a stub.
                    cmiReferenceable = getCMIReferenceable(cmiReference);
                    logger.debug("Object: {0} - Invoke: {1} > CMIReferenceable successfully retrieved: {2}",
                            objectName, methodName, cmiReferenceable);

                    decisionManagerOnChoose = policy.onChoose(method, args, cmiReferenceable);

                } catch (Exception e) {
                    logger.debug("Object:{0} - Invoke: {1} > onLookupException.", objectName, methodName, e);

                    // Check if the policy allows failover
                    DecisionManager<Void> decisionManager = policy.onLookupException(cmiReference, e);
                    if(decisionManager.getDecision().equals(Decision.THROW)) {
                        // Failover is not allowed, rethrow the exception
                        logger.debug("Object: {0} - Invoke: {1} > onLookupException=THROW",
                                objectName, methodName, decisionManager.getThrowable());
                        throw decisionManager.getThrowable();
                    }

                    // Failover is allowed, delete the last reference
                    logger.debug("Object: {0} - Invoke: {1} > onLookupException=RETRY, Removing node {2}",
                            objectName, methodName, cmiReference);
                    cmiReferences.remove(cmiReference);
                    currentRef = null;

                    // Retry one more time...
                    continue;
                }

                // Take a decision concerning the retrieved delegate
                Decision decision = decisionManagerOnChoose.getDecision();
                if(decision.equals(Decision.RETRY)) {
                    // Choose an other delegate
                    logger.debug("Object:{0} - Invoke: {1} > onChoose=RETRY!", objectName, method);
                    cmiReferences.remove(cmiReference);
                    continue;
                } else if(decision.equals(Decision.RETURN)) {
                    // OK: We can now perform the invocation
                    logger.debug("Object: {0} - Invoke: {1} > onChoose=RETURN!", objectName, method);
                    // Keep the delegate
                    cmiReferenceable = decisionManagerOnChoose.getRetVal();
                } else if(decision.equals(Decision.THROW)) {
                    // KO: Throw an exception
                    logger.error("Object: {0} - Invoke: {1} > onChoose=THROW!",
                            objectName, method, decisionManagerOnChoose.getThrowable());
                    throw decisionManagerOnChoose.getThrowable();
                }

                // Keep a trace of this mandatory
                currentRef = cmiReferenceable;

            }

            if(cmiReferenceable == null) {
                // Invalid state
                logger.error("Object: {0} - Invoke: {1} > The current manadatory for reference {2} should not be null !",
                        objectName, methodName, cmiReference);
                throw new CMIInvocationHandlerException(
                        "Object: " + objectName + " - Invoke: " + methodName
                        + " > The current manadatory for reference " + cmiReference + "should not be null !");
            }

            // Callback for extending CMIInvocationHandler
            preInvokeHook(cmiReferenceable);

            ClassLoader oldClassLoader = null;

            // Use the classloader dedicated to the invocation
            if(classLoader != null) {
                oldClassLoader = Thread.currentThread().getContextClassLoader();
                Thread.currentThread().setContextClassLoader(classLoader);
            }

            logger.debug("Object: {0} - Invoke: {1} > Invocation with args: {2}", objectName, methodName, args);

            try {
                // Perform the invocation
                Object result;
                long timeBefore = System.currentTimeMillis();
                try {
                    result = method.invoke(cmiReferenceable.getReferencedObject(), args);
                    long timeAfter = System.currentTimeMillis();
                    lastInvocDuration = timeAfter - timeBefore;
                    logger.debug("Object: {0} - Invoke: {1} > Invocation performed in {2} ms",
                            objectName, methodName, lastInvocDuration);
                } finally {
                    // Reuse the previous classloader
                    if(classLoader != null) {
                        Thread.currentThread().setContextClassLoader(oldClassLoader);
                    }
                }
                // Callback for extending CMIInvocationHandler
                postInvokeHook();

                // Use the callback of policy before returning the value
                decisionManagerOnReturn = policy.onReturn(method, args, cmiReference, result);

            } catch(InvocationTargetException e) {
                logger.debug("Object: {0} - Invoke: {1} > onInvokeException.", objectName, methodName, e.getCause());
                // Check if the policy allows failover
                DecisionManager<Void> decisionManager =
                    policy.onInvokeException(method, args, cmiReference, e.getCause());

                if(decisionManager.getDecision().equals(Decision.THROW)) {
                    // Failover is not allowed, rethrow the exception
                    logger.debug("Object: {0} - Invoke: {1} > onInvokeException=THROW!",
                            objectName, methodName, decisionManager.getThrowable());
                    throw decisionManager.getThrowable();
                }

                if(clusterViewManager instanceof ClientClusterViewManager && !itf.equals(ClientClusterViewProvider.class)) {

                    // Force a update of the node to avoid a reconnection on this bad server.
                    synchronized (this) {
                        logger.debug("Object: {0} - Invoke: {1} > onInvokeException=RETRY, Forcing an update...",
                                objectName, methodName);
                        ((ClientClusterViewManager) clusterViewManager).pullAndUpdateObjectInfos(objectName);
                    }
                }

                cmiReferences =
                    new ArrayList<CMIReference>(clusterViewManager.getCMIReferences(objectName, protocolName));
                logger.debug("Object: {0} - Invoke: {1} > onInvokeException=RETRY, CMIReferences is now: {2}",
                        objectName, methodName, cmiReferences);

                // Delete the last reference
                logger.debug("Object: {0} - Invoke: {1} > onInvokeException=RETRY, Removing node {2}",
                        objectName, methodName, cmiReference);
                cmiReferences.remove(cmiReference);

                currentRef = null;

                // Callback for extending CMIInvocationHandler
                onExceptionHook(objectName, cmiReferenceable);
            } catch(Exception e) {
                logger.error("Object: {0} - Invoke: {1} > onInvokeException.", objectName, methodName, e);
                throw new CMIInvocationHandlerException(
                        "Object: " + objectName + " - Invoke: " + method
                        + " - Args: " + Arrays.toString(args) + " - Arg types: " + getArgTypes(args)
                        + " > onInvokeException.", e);
            } finally {
                // Callback for extending CMIInvocationHandler
                onFinallyHook(objectName, cmiReferenceable);
            }

            // Take a decision concerning the returned value
            if(decisionManagerOnReturn != null) {
                Decision decision = decisionManagerOnReturn.getDecision();
                if(decision.equals(Decision.RETRY)) {
                    logger.debug("Object: {0} - Invoke: {1} > onReturn=RETRY!", objectName, method);
                    cmiReferences.remove(cmiReference);
                        currentRef = null;
                        decisionManagerOnReturn = null;
                } else if(decision.equals(Decision.RETURN)) {
                    logger.debug("Object: {0} - Invoke: {1} > onReturn=RETURN!", objectName, method);
                } else if(decision.equals(Decision.THROW)) {
                    logger.error("Object: {0} - Invoke: {1} > onReturn=THROW!",
                        objectName, method, decisionManagerOnReturn.getThrowable());
                    throw decisionManagerOnReturn.getThrowable();
                }
            }
        } while(decisionManagerOnReturn == null);

        if(cmiProxyHandle != null) {
            cmiProxyHandle = cmiProxyHandle.updateHttpSession();
        }
        return decisionManagerOnReturn.getRetVal();
    }

    /**
     * @param args arguments
     * @return the string of the argument class names
     */
    private String getArgTypes(final Object... args) {
        if (args == null) {
            return "null";
        }
        List<String> classnames = new ArrayList<String>(args.length);
        for (Object arg : args) {
            classnames.add(arg.getClass().getName());
        }
        return classnames.toString();
    }

    /**
     * @return the object to use for the invocation
     */
    public CMIReferenceable<T> getCurrentRef() {
        return currentRef;
    }

    /**
     * Set the object to use for the invocation.
     * @param currentRef the object to use for the invocation
     */
    public void setCurrentRef(final CMIReferenceable<T> currentRef) {
        this.currentRef = currentRef;
    }

    /**
     * @return the method name to get an handle
     */
    protected String getHandleMethodName() {
        return "getHandle";
    }

    /**
     * @param cmiProxy a cmi proxy
     * @return an handle for the given proxy
     */
    protected abstract CMIProxyHandle getHandle(CMIProxy cmiProxy);

    /**
     * @param cmiProxyHandle
     */
    public void setCmiProxyHandle(final CMIProxyHandle cmiProxyHandle) {
        this.cmiProxyHandle = cmiProxyHandle;
    }

    /**
     * Get a CMIreferenceable for a given CMIReference.
     * @param cmiReference
     * @return
     * @throws Exception
     */
    protected abstract CMIReferenceable<T> getCMIReferenceable(CMIReference cmiReference) throws Exception;

    /**
     * @return duration of the last invocation
     */
    public long getLastInvocDuration() {
        return lastInvocDuration;
    }

    /*********************** Begin of CMIInvocationHandler callbacks ***********************/

    /**
     * Preinvocation callback.
     * @param cmiReferenceable the mandatory that will be used for the invocation
     */
    protected void preInvokeHook(final CMIReferenceable<T> cmiReferenceable){}

    /**
     * Postinvocation callback.
     */
    protected void postInvokeHook(){}

    /**
     * Onexception callback.
     * @param objectName
     * @param cmiReferenceable
     * @throws Exception
     */
    protected abstract void onExceptionHook(String objectName, CMIReferenceable<T> cmiReferenceable) throws Throwable;

    /**
     * Finally callback.
     * @param objectName
     * @param cmiReferenceable
     */
    protected abstract void onFinallyHook(String objectName, CMIReferenceable<T> cmiReferenceable);


    /*********************** End of CMIInvocationHandler callbacks ***********************/

}
